/*
 * Copyright (C) 2010 Google Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1.  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 * 2.  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 * 3.  Neither the name of Apple Computer, Inc. ("Apple") nor the names of
 *     its contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY APPLE AND ITS CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL APPLE OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef AudioListener_h
#define AudioListener_h

#include "modules/webaudio/AudioParam.h"
#include "platform/bindings/ScriptWrappable.h"
#include "platform/geometry/FloatPoint3D.h"
#include "platform/heap/Handle.h"

namespace blink {

class HRTFDatabaseLoader;
class PannerHandler;

// AudioListener maintains the state of the listener in the audio scene as
// defined in the OpenAL specification.

class AudioListener : public ScriptWrappable {
  DEFINE_WRAPPERTYPEINFO();

 public:
  static AudioListener* Create(BaseAudioContext& context) {
    return new AudioListener(context);
  }
  virtual ~AudioListener();

  // Location of the listener
  AudioParam* positionX() const { return position_x_; };
  AudioParam* positionY() const { return position_y_; };
  AudioParam* positionZ() const { return position_z_; };

  // Forward direction vector of the listener
  AudioParam* forwardX() const { return forward_x_; };
  AudioParam* forwardY() const { return forward_y_; };
  AudioParam* forwardZ() const { return forward_z_; };

  // Up direction vector for the listener
  AudioParam* upX() const { return up_x_; };
  AudioParam* upY() const { return up_y_; };
  AudioParam* upZ() const { return up_z_; };

  // True if any of AudioParams have automations.
  bool HasSampleAccurateValues() const;

  // Update the internal state of the listener, including updating the dirty
  // state of all PannerNodes if necessary.
  void UpdateState();

  bool IsListenerDirty() const { return is_listener_dirty_; }

  const FloatPoint3D GetPosition() const {
    return FloatPoint3D(position_x_->value(), position_y_->value(),
                        position_z_->value());
  }
  const FloatPoint3D Orientation() const {
    return FloatPoint3D(forward_x_->value(), forward_y_->value(),
                        forward_z_->value());
  }
  const FloatPoint3D UpVector() const {
    return FloatPoint3D(up_x_->value(), up_y_->value(), up_z_->value());
  }

  const float* GetPositionXValues(size_t frames_to_process);
  const float* GetPositionYValues(size_t frames_to_process);
  const float* GetPositionZValues(size_t frames_to_process);

  const float* GetForwardXValues(size_t frames_to_process);
  const float* GetForwardYValues(size_t frames_to_process);
  const float* GetForwardZValues(size_t frames_to_process);

  const float* GetUpXValues(size_t frames_to_process);
  const float* GetUpYValues(size_t frames_to_process);
  const float* GetUpZValues(size_t frames_to_process);

  // Position
  void setPosition(float x, float y, float z) {
    setPosition(FloatPoint3D(x, y, z));
  }

  // Orientation and Up-vector
  void setOrientation(float x,
                      float y,
                      float z,
                      float up_x,
                      float up_y,
                      float up_z) {
    setOrientation(FloatPoint3D(x, y, z));
    SetUpVector(FloatPoint3D(up_x, up_y, up_z));
  }

  Mutex& ListenerLock() { return listener_lock_; }
  void AddPanner(PannerHandler&);
  void RemovePanner(PannerHandler&);

  // HRTF DB loader
  HRTFDatabaseLoader* HrtfDatabaseLoader() {
    return hrtf_database_loader_.get();
  }
  void CreateAndLoadHRTFDatabaseLoader(float);
  bool IsHRTFDatabaseLoaded();
  void WaitForHRTFDatabaseLoaderThreadCompletion();

  void Trace(blink::Visitor*);

 private:
  AudioListener(BaseAudioContext&);

  void setPosition(const FloatPoint3D&);
  void setOrientation(const FloatPoint3D&);
  void SetUpVector(const FloatPoint3D&);

  void MarkPannersAsDirty(unsigned);

  // Location of the listener
  Member<AudioParam> position_x_;
  Member<AudioParam> position_y_;
  Member<AudioParam> position_z_;

  // Forward direction vector of the listener
  Member<AudioParam> forward_x_;
  Member<AudioParam> forward_y_;
  Member<AudioParam> forward_z_;

  // Up direction vector for the listener
  Member<AudioParam> up_x_;
  Member<AudioParam> up_y_;
  Member<AudioParam> up_z_;

  // The position, forward, and up vectors from the last rendering quantum.
  FloatPoint3D last_position_;
  FloatPoint3D last_forward_;
  FloatPoint3D last_up_;

  // Last time that the automations were updated.
  double last_update_time_;

  // Set every rendering quantum if the listener has moved in any way
  // (position, forward, or up).  This should only be read or written to from
  // the audio thread.
  bool is_listener_dirty_;

  void UpdateValuesIfNeeded(size_t frames_to_process);

  AudioFloatArray position_x_values_;
  AudioFloatArray position_y_values_;
  AudioFloatArray position_z_values_;

  AudioFloatArray forward_x_values_;
  AudioFloatArray forward_y_values_;
  AudioFloatArray forward_z_values_;

  AudioFloatArray up_x_values_;
  AudioFloatArray up_y_values_;
  AudioFloatArray up_z_values_;

  // Synchronize a panner's process() with setting of the state of the listener.
  mutable Mutex listener_lock_;
  // List for pannerNodes in context. This is updated only in the main thread,
  // and can be referred in audio thread.
  // These raw pointers are safe because PannerHandler::uninitialize()
  // unregisters it from m_panners.
  HashSet<PannerHandler*> panners_;
  // HRTF DB loader for panner node.
  scoped_refptr<HRTFDatabaseLoader> hrtf_database_loader_;
};

}  // namespace blink

#endif  // AudioListener_h
