/*
 * Copyright (C) 2010 Google Inc. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above
 * copyright notice, this list of conditions and the following disclaimer
 * in the documentation and/or other materials provided with the
 * distribution.
 *     * Neither the name of Google Inc. nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef AudioDSPKernel_h
#define AudioDSPKernel_h

#include "platform/audio/AudioDSPKernelProcessor.h"
#include "platform/wtf/Allocator.h"

namespace blink {

// AudioDSPKernel does the processing for one channel of an
// AudioDSPKernelProcessor.

class PLATFORM_EXPORT AudioDSPKernel {
  USING_FAST_MALLOC(AudioDSPKernel);

 public:
  AudioDSPKernel(AudioDSPKernelProcessor* kernel_processor)
      : kernel_processor_(kernel_processor),
        sample_rate_(kernel_processor->SampleRate()) {}

  AudioDSPKernel(float sample_rate)
      : kernel_processor_(nullptr), sample_rate_(sample_rate) {}

  virtual ~AudioDSPKernel();

  // Subclasses must override process() to do the processing and reset() to
  // reset DSP state.
  virtual void Process(const float* source,
                       float* destination,
                       size_t frames_to_process) = 0;
  // Subclasses that have AudioParams must override this to process the
  // AudioParams.
  virtual void ProcessOnlyAudioParams(size_t frames_to_process){};
  virtual void Reset() = 0;

  float SampleRate() const { return sample_rate_; }
  double Nyquist() const { return 0.5 * SampleRate(); }

  AudioDSPKernelProcessor* Processor() { return kernel_processor_; }
  const AudioDSPKernelProcessor* Processor() const { return kernel_processor_; }

  virtual double TailTime() const = 0;
  virtual double LatencyTime() const = 0;

 protected:
  // This raw pointer is safe because the AudioDSPKernelProcessor object is
  // guaranteed to be kept alive while the AudioDSPKernel object is alive.
  AudioDSPKernelProcessor* kernel_processor_;
  float sample_rate_;
};

}  // namespace blink

#endif  // AudioDSPKernel_h
